/* $Id: addressbook.C,v 1.3 2003/09/15 04:27:13 mrsam Exp $
**
** Copyright 2003, Double Precision Inc.
**
** See COPYING for distribution information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include "curses/cursesscreen.H"
#include "curses/cursesbutton.H"
#include "curses/cursestitlebar.H"
#include "curses/cursesstatusbar.H"
#include "curses/cursesmainscreen.H"
#include "curses/cursesdialog.H"
#include "curses/curseslabel.H"
#include "curses/cursesfield.H"
#include "curses/curseskeyhandler.H"

#include "gettext.H"
#include "myserver.H"
#include "globalkeys.H"
#include "myservercallback.H"
#include "myserverpromptinfo.H"
#include "myserverlogincallback.H"
#include "previousscreen.H"
#include "passwordlist.H"

#include "libmail/rfcaddr.H"

#include "opensubfolders.H"
#include "addressbook.H"

#include "libmail/logininfo.H"
#include "libmail/rfc2047decode.H"

#include <vector>
#include <list>
#include <set>
#include <map>

using namespace std;

extern Gettext::Key key_DELADDRESSBOOKENTRY;
extern Gettext::Key key_ADDADDRESSBOOK;
extern Gettext::Key key_DELETEADDRESSBOOK;
extern Gettext::Key key_ABORT;
extern Gettext::Key key_UP;
extern Gettext::Key key_DOWN;

extern CursesScreen *cursesScreen;
extern CursesMainScreen *mainScreen;
extern CursesStatusBar *statusBar;
extern CursesTitleBar *titleBar;

extern void mainMenu(void *);

void addressbookIndexScreen(void *);

list<AddressBook *> AddressBook::addressBookList;
string AddressBook::importName;
string AddressBook::importAddr;

AddressBook *AddressBook::defaultAddressBook=NULL;

AddressBook::AddressBook()
	: server(NULL),
	  folder(NULL),
	  addressBook(NULL),
	  iwantFocus(false)
{
}

AddressBook::~AddressBook()
{
	if (addressBook)
	{
		delete addressBook;
		addressBook=NULL;
	}

	if (folder)
	{
		delete folder;
		folder=NULL;
	}

	mail::account *s=server;

	if (s)
	{
		server=NULL;

		myServer::Callback logout_callback;

		s->logout(logout_callback);

		myServer::eventloop(logout_callback);
		delete server;
	}
}

void AddressBook::disconnected(const char *errmsg)
{
	if (errmsg && *errmsg && server)
	{
		statusBar->status(errmsg, statusBar->DISCONNECTED);
		statusBar->beepError();
	}

	if (folder)
	{
		delete folder;
		folder=NULL;
	}

	if (server)
	{
		mail::account *serverPtr=server;
		server=NULL;
		delete serverPtr;
	}
}

void AddressBook::servererror(const char *errmsg)
{
	if (errmsg && *errmsg)
	{
		statusBar->status(errmsg, statusBar->SERVERERROR);
		statusBar->beepError();
	}

	if (folder)
	{
		delete folder;
		folder=NULL;
	}

	if (server)
	{
		delete server;
		server=NULL;
	}
}

void AddressBook::init()
{
	// Open the default address book first.

	string afile=myServer::getConfigDir()+ "/addressbook";

	int fd= ::open(afile.c_str(), O_RDWR|O_CREAT, 0600);

	if (fd >= 0)
		::close(fd);

	if (defaultAddressBook)
		return;

	defaultAddressBook=new AddressBook;

	if (!defaultAddressBook)
		outofmemory();

	try {
		string url="mbox:" + afile;

		myServer::Callback login_callback;

		mail::account::openInfo loginInfo;

		loginInfo.url=url;

		mail::account *abook=mail::account::open(loginInfo,
							 login_callback,
							 *defaultAddressBook);

		if (!abook)
			outofmemory();

		// Need to do some song-n-dance with LibMAIL in order to
		// initialize the default address book folder.

		try {
			if (!myServer::eventloop(login_callback))
			{
				delete abook;
				delete defaultAddressBook;
				return;
			}

			OpenSubFoldersCallback readFolders;
			myServer::Callback readfolders_callback;

			abook->readTopLevelFolders(readFolders,
						   readfolders_callback);
			if (!myServer::eventloop(login_callback))
			{
				delete abook;
				delete defaultAddressBook;
				return;
			}

			if (readFolders.folders.size() > 0)
			{
				defaultAddressBook->init(_("Local Address Book"),
							 url,
							 readFolders
							 .folders[0]
							 ->toString());

				myServer::Callback logout_callback;

				abook->logout(logout_callback);
				myServer::eventloop(logout_callback);

				delete abook;
				abook=NULL;
			}
			else
			{
				delete defaultAddressBook;
				defaultAddressBook=NULL;
			}

		} catch (...) {
			if (abook)
				delete abook;
			LIBMAIL_THROW();
		}
	} catch (...) {
		delete defaultAddressBook;
		defaultAddressBook=NULL;
		LIBMAIL_THROW();
	}
}

void AddressBook::init(string nameArg,
		       string urlArg,
		       string folderStrArg)
{
	name=nameArg;
	url=urlArg;
	folderStr=folderStrArg;
}

bool AddressBook::open()
{
	if (addressBook)
		return true; // Already open.

	if (folder)
	{
		delete folder;
		folder=NULL;
	}

	if (server)
	{
		mail::account *s=server;

		server=NULL;
		delete s;
	}

	// Log in.

	string pwd="";

	if (!PasswordList::passwordList.check(url, pwd) ||
	    !serverLogin(pwd, NULL))
	{
		PasswordList::passwordList.remove(url);

		myServerLoginCallback loginCallback;

		if (!serverLogin("", &loginCallback))
			return false;
	}

	// Build the folder object, then open it.

	folder=server->folderFromString(folderStr);

	if (!folder)
	{
		servererror(strerror(errno));
		return false;
	}

	addressBook=new mail::addressbook(server, folder);

	if (!addressBook)
		outofmemory();

	myServer::Callback selectCallback;
	addressBook->open(selectCallback);
	if (!myServer::eventloop(selectCallback))
	{
		delete addressBook;
		addressBook=NULL;
	}
	return addressBook != NULL;
}

// Try to log in to an addressbook account.

bool AddressBook::serverLogin(string pwd, myServerLoginCallback *loginPrompt)
{
	myServer::Callback login_callback;

	if (loginPrompt)
	{
		loginPrompt->myCallback= &login_callback;
	}

	mail::account::openInfo loginInfo;
	loginInfo.url=url;
	loginInfo.pwd=pwd;
	loginInfo.loginCallbackObj=loginPrompt;

	server=mail::account::open(loginInfo,
				   login_callback, *this);

	if (!server)
		outofmemory();


	for (;;)
	{
		if (loginPrompt)
		{
			loginPrompt->reset();
			loginPrompt->myCallback= &login_callback;
		}

		statusBar->clearstatus();
		statusBar->status(Gettext(_("Connecting to %1%...")) << name);

		bool rc=myServer::eventloop(login_callback);

		if (loginPrompt && login_callback.interrupted)
		{
			// Using a login callback (instead of verbatim passwd)

			string promptStr=
				Gettext( loginPrompt->isPasswordPrompt
					 ?_("Password for %1% (CTRL-C - cancel): ")
					 :_("Login for %1% (CTRL-C - cancel): ")
					 ) << name;

			myServer::promptInfo response=
				myServer::promptInfo(promptStr);

			if (loginPrompt->isPasswordPrompt)
				response.password();

			response=myServer::prompt(response);

			if (response.aborted())
			{
				loginPrompt->callbackCancel();
				continue;
			}

			pwd=response;
			loginPrompt->callback(pwd);
			continue;
		}

		if (!rc)
		{
			if (server)
			{
				mail::account *serverPtr=server;
				server=NULL;
				delete serverPtr;
			}
			return false;
		}
		break;
	}

	// Make sure we cache the good password.
	PasswordList::passwordList.save(url, pwd);
	return true;
}


void AddressBook::closeAllRemote()
{
	list<AddressBook *>::iterator b, e;

	b=addressBookList.begin();
	e=addressBookList.end();

	while (b != e)
	{
		AddressBook *p= *b;

		p->close();
		b++;
	}
}

void AddressBook::closeAll()
{
	if (defaultAddressBook)
	{
		defaultAddressBook->close();
		delete defaultAddressBook;
		defaultAddressBook=NULL;
	}

	list<AddressBook *>::iterator b, e;

	while ((b=addressBookList.begin()) != (e=addressBookList.end()))
	{
		AddressBook *p= *b;

		addressBookList.erase(b);

		try {
			p->close();
			delete p;
		} catch (...) {
			delete p;
		}
	}
}

void AddressBook::close()
{
	mail::account *serverPtr=server;

	// Be polite and formally log off the server

	if (serverPtr)
		try {
			server=NULL;

			myServer::Callback logout_callback;

			serverPtr->logout(logout_callback);

			myServer::eventloop(logout_callback);

			delete serverPtr;
			if (addressBook)
			{
				delete addressBook;
				addressBook=NULL;
			}

			if (folder)
			{
				delete folder;
				folder=NULL;
			}

		} catch (...) {
			delete serverPtr;
			if (addressBook)
			{
				delete addressBook;
				addressBook=NULL;
			}

			if (folder)
			{
				delete folder;
				folder=NULL;
			}
			LIBMAIL_THROW();
		}
}

// Add a new address book entry.

bool AddressBook::add(mail::addressbook::Entry &newEntry)
{
	if (!open())
		return false;

	myServer::Callback add_callback;

	addressBook->add(newEntry,
			 currentRecord.second, // Old UID, if not empty
			 add_callback);

	return myServer::eventloop(add_callback);
}

// Delete an address book entry

bool AddressBook::del(string uid)
{
	if (!open())
		return false;

	myServer::Callback del_callback;

	addressBook->del(uid, del_callback);

	return myServer::eventloop(del_callback);
}

// Search the address book for a specific nickname.
bool AddressBook::search(std::string nickname,
			 std::vector<mail::address> &addrListArg)
{
	if (!open())
		return false;

	myServer::Callback search_callback;

	addressBook->search(nickname, addrListArg, search_callback);

	return myServer::eventloop(search_callback);
}

// List address book contents.

void AddressBook::getIndex( list< pair<string, string> > &listArg)
{
	if (!open())
		return;

	myServer::Callback getindex_callback;

	addressBook->getIndex(listArg, getindex_callback);

	myServer::eventloop(getindex_callback);
}

//
// Read the current address book entry into addrList:

bool AddressBook::getAddressList(vector<mail::address> &addrList)
{
	myServer::Callback get_callback;

	addressBook->getAddressList(currentRecord.second,
				    addrList, get_callback);
	return myServer::eventloop(get_callback);
}

//
// Replace addresses in the address list with anything found in
// the address books.

bool AddressBook::searchAll(vector<mail::address> &addrList, // Orig list
			    vector<mail::address> &retList) // Final list
{
	set<string> addrSet; // Eliminate dupes

	vector<mail::address>::iterator curAddr=addrList.begin(),
		lastAddr=addrList.end();

	for ( ; curAddr != lastAddr; curAddr++)
	{
		if (curAddr->addr.find('@') != curAddr->addr.npos)
		{
			// Not a nickname, a dupe?

			if (addrSet.count( curAddr->addr ) == 0)
			{
				addrSet.insert(curAddr->addr);
				retList.push_back( *curAddr);
			}
			continue;
		}

		// Go through each address book, see what we can find.

		list<AddressBook *>::iterator ab=addressBookList.begin(),
			ae=addressBookList.end();

		vector<mail::address> addresses;

		if (defaultAddressBook &&
		    !defaultAddressBook->search(curAddr->addr, addresses))
			return false;

		while (ab != ae && addresses.size() == 0)
		{
			if (! (*ab++)->search(curAddr->addr, addresses))
				return false;
		}

		if (addresses.size() == 0) // Not found, check the pw file
		{
			mail::address addr= *curAddr;

			if (addr.name.size() == 0)
			{
				struct passwd *pw=getpwnam(addr.addr.c_str());

				if (pw != NULL)
				{
					addr.name=pw->pw_gecos;

					size_t n=addr.name.find(',');

					if (n != addr.name.npos)
						addr.name=addr.name
							.substr(0, n);
				}
			}

			addr.addr += "@";
			addr.addr += mail::account::hostname();

			addresses.push_back(addr);
		}

		// Found addresses, check for dupes, add them.

		vector<mail::address>::iterator addrb=addresses.begin(),
			addre=addresses.end();

		while (addrb != addre)
		{
			if (addrSet.count( addrb->addr ) == 0)
			{
				addrSet.insert(addrb->addr);
				retList.push_back( *addrb );
			}

			++addrb;
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////
//
// Add new address book entry screen.

class AddAddressBookScreen : public CursesContainer,
			     public CursesKeyHandler {

	AddressBook *addressBook;
	CursesLabel title;
	CursesDialog addDialog;

	CursesLabel nickname_label;
	CursesLabel name_label;
	CursesLabel addr_label;

	CursesField nickname_field;
	CursesField name_field;
	CursesField addr_field;

	CursesButtonRedirect<AddAddressBookScreen> saveButton, cancelButton;

	void save();
	void cancel();

	bool processKey(const Curses::Key &key);

public:
	void requestFocus()
	{
		nickname_field.requestFocus();
	}

	AddAddressBookScreen(CursesMainScreen *,
			     AddressBook *);
	~AddAddressBookScreen();
};

AddAddressBookScreen::AddAddressBookScreen(CursesMainScreen *mainScreen,
					   AddressBook *addressBookArg)
	: CursesContainer(mainScreen),
	  CursesKeyHandler(PRI_SCREENHANDLER),
	  addressBook(addressBookArg),
	  title(this, _("Add Entry")),
	  addDialog(this),
	  nickname_label(NULL, _("Nickname: ")),
	  name_label(NULL, _("Name: ")),
	  addr_label(NULL, _("Address: ")),

	  nickname_field(NULL),
	  name_field(NULL),
	  addr_field(NULL),

	  saveButton(this, _("SAVE")),
	  cancelButton(this, _("CANCEL"))
{
	setRow(4);
	setAlignment(Curses::PARENTCENTER);

	title.setAlignment(Curses::PARENTCENTER);

	addDialog.setRow(2);

	addDialog.addPrompt(&nickname_label, &nickname_field, 0);
	addDialog.addPrompt(&name_label, &name_field, 1);
	addDialog.addPrompt(&addr_label, &addr_field, 2);

	titleBar->setTitles(_("ADD ENTRY"), "");

	saveButton=this;
	cancelButton=this;

	saveButton= &AddAddressBookScreen::save;
	cancelButton= &AddAddressBookScreen::cancel;

	addDialog.addPrompt(NULL, &saveButton, 4);
	addDialog.addPrompt(NULL, &cancelButton, 6);

	vector<mail::address> addresses;

	if (addressBook->currentRecord.second.size() > 0) // Edit exist rec
	{
		nickname_field.setText(addressBook->currentRecord.first);

		if (!addressBook->getAddressList(addresses))
			addresses.clear();
	}

	// Get any defaults from the take addresses screen.

	if (AddressBook::importAddr.size() > 0)
		addresses.push_back( mail::address(AddressBook::importName,
						   AddressBook::importAddr));

	if (addresses.size() > 0)
	{
		if (addresses.size() == 1)
		{
			name_field.setText(addresses[0].name);
			addr_field.setText(addresses[0].addr);
		}
		else // It's a list.
		{
			string s=mail::address::toString("", addresses, 0);

			string::iterator b=s.begin(), e=s.end();

			while (b != e)
			{
				if (*b == '\n')
					*b=' ';
				b++;
			}

			addr_field.setText(s);
		}
	}

	name_field.requestFocus();
}

bool AddAddressBookScreen::processKey(const Curses::Key &key)
{
	if (key == key_ABORT)
	{
		keepgoing=false;
		myServer::nextScreen= &mainMenu;
		myServer::nextScreenArg=NULL;
		PreviousScreen::previousScreen();
		return true;
	}
	return false;
}

AddAddressBookScreen::~AddAddressBookScreen()
{
}

void AddAddressBookScreen::save()
{
	mail::addressbook::Entry addEntry;

	addEntry.nickname=nickname_field.getText();
	addEntry.name=name_field.getText();
	addEntry.address=addr_field.getText();

	if (addEntry.nickname.size() == 0)
	{
		nickname_field.requestFocus();
		nickname_field.beepError();
		return;
	}

	if (addEntry.address.size() == 0)
	{
		addr_field.requestFocus();
		addr_field.beepError();
		return;
	}

	if (addressBook->add(addEntry))
	{
		keepgoing=false;
		myServer::nextScreen= &addressbookIndexScreen;
		myServer::nextScreenArg=addressBook;
		return;
	}
}

void AddAddressBookScreen::cancel()
{
	keepgoing=false;
	myServer::nextScreen= &mainMenu;
	myServer::nextScreenArg=NULL;
	PreviousScreen::previousScreen();
}

void addressbookScreen(void *vp)
{
	AddressBook *a=(AddressBook *)vp;

	AddAddressBookScreen screen(mainScreen, a);

	screen.requestFocus();
	myServer::eventloop();
}

/////////////////////////////////////////////////////////////////////////////
//
// Address book Index
//

class AddressBookIndexScreen : public CursesContainer,
			       public CursesKeyHandler {

	AddressBook *addressBook;
	CursesLabel title;
	CursesDialog indexDialog;

	class indexButton : public CursesButton {
		AddressBookIndexScreen *myScreen;
		pair<string, string> myEntry;
		// first: nickname, second: uid

		bool processKeyInFocus(const Key &key);

	public:
		size_t indexNum;

		indexButton(AddressBookIndexScreen *myScreenArg,
			    pair<string, string> myEntryArg);
		~indexButton();

		void clicked();
	};

	vector<indexButton *> buttons;

	class indexSort {
	public:
		bool operator()(const pair<string, string> &a,
				const pair<string, string> &b);
	};

	bool processKey(const Curses::Key &key);
	bool listKeys( vector< pair<string,  string> > &list);
public:
	void requestFocus()
	{
		if (buttons.size() > 0)
			buttons[0]->requestFocus();
	}

	AddressBookIndexScreen(CursesMainScreen *, AddressBook *,
			       vector< pair<string, string> > &index);
	~AddressBookIndexScreen();

	void clicked( pair<string, string> entryArg);
	void deleteEntry(size_t n, string uid);
};

AddressBookIndexScreen::indexButton::indexButton(AddressBookIndexScreen
						 *myScreenArg,
						 pair<string, string>
						 myEntryArg)
	: CursesButton(myScreenArg, myEntryArg.first),
	  myScreen(myScreenArg),
	  myEntry(myEntryArg),
	  indexNum(0)
{
	setStyle(MENU);
}

AddressBookIndexScreen::indexButton::~indexButton()
{
}

void AddressBookIndexScreen::indexButton::clicked()
{
	myScreen->clicked(myEntry);
}

bool AddressBookIndexScreen::indexButton::processKeyInFocus(const Key &key)
{
	if (key == key_DELADDRESSBOOKENTRY)
	{
		myServer::promptInfo response=
			myServer
			::promptInfo(Gettext(_("Delete address book entry for '%1%'? (Y/N) "))
				     << myEntry.first).yesno();

		response=myServer::prompt(response);

		if (response.aborted())
			return true;

		if ((string)response == "Y")
			myScreen->deleteEntry(indexNum, myEntry.second);

		return true;
	}
	return CursesButton::processKeyInFocus(key);
}

AddressBookIndexScreen::AddressBookIndexScreen(CursesMainScreen *parent,
					       AddressBook *addressBookArg,
					       vector< pair<string, string> >
					       &indexArg)
	: CursesContainer(parent),
	  CursesKeyHandler(PRI_SCREENHANDLER),
	  addressBook(addressBookArg),
	  title(this, addressBookArg->getName()),
	  indexDialog(this)
{
	setRow(4);
	setAlignment(Curses::PARENTCENTER);
	title.setAlignment(Curses::PARENTCENTER);

	indexDialog.setRow(2);
	titleBar->setTitles(_("ADDRESS BOOK INDEX"), "");

	sort(indexArg.begin(), indexArg.end(), indexSort());

	buttons.reserve(indexArg.size());
	vector< pair<string, string> >::iterator b=indexArg.begin(),
		e=indexArg.end();

	while (b != e)
	{
		indexButton *bb=new indexButton(this, *b);
		b++;

		if (!bb)
			outofmemory();

		try {
			bb->indexNum=buttons.size();
			buttons.insert(buttons.end(), bb);
		} catch (...) {
			delete bb;
			LIBMAIL_THROW();
		}

		indexDialog.addPrompt(NULL, bb);
	}
}

void AddressBookIndexScreen::deleteEntry(size_t n, string uid)
{
	if (addressBook->del(uid))
	{
		indexButton *b=buttons[n];

		buttons.erase(buttons.begin() + n);

		delete b;

		vector<indexButton *>::iterator p=buttons.begin() + n,
			e=buttons.end();

		if (p != e)
			(*p)->requestFocus();
		else if (buttons.size() > 0)
			(*(buttons.end()-1))->requestFocus();

		while (p != e)
		{
			(*p)->indexNum=n++;
			(*p)->setRow( (*p)->getRow() - 1);

			p++;
		}
	}
}

bool AddressBookIndexScreen
::indexSort::operator()(const pair<string, string> &a,
			const pair<string, string> &b)
{
	return (a.first.compare(b.first) < 0);
}

bool AddressBookIndexScreen::processKey(const Curses::Key &key)
{
	if (key == key_ABORT)
	{
		keepgoing=false;
		myServer::nextScreen= &mainMenu;
		myServer::nextScreenArg=NULL;
		PreviousScreen::previousScreen();
		return true;
	}

	if (key == key_ADDADDRESSBOOK)
	{
		keepgoing=false;
		addressBook->currentRecord.second="";
		myServer::nextScreen= &addressbookScreen;
		myServer::nextScreenArg=addressBook;
		return true;
	}

	return GlobalKeys::processKey(key, GlobalKeys::ADDRESSBOOKINDEXSCREEN,
				      NULL);
}

bool AddressBookIndexScreen::listKeys( vector< pair<string,  string> > &list)
{
	GlobalKeys::listKeys(list, GlobalKeys::ADDRESSBOOKINDEXSCREEN);

	list.push_back(make_pair(Gettext::keyname(_("ADDADDRESSBOOK_K:A")),
				 _("Add")));
	list.push_back(make_pair(Gettext::keyname(_("DELADDRESSBOOKENTRY_K:D"))
				 , _("Delete")));
	return false;
}

AddressBookIndexScreen::~AddressBookIndexScreen()
{
	vector<indexButton *>::iterator b=buttons.begin(), e=buttons.end();

	while (b != e)
	{
		indexButton *p= *b;

		*b++ = NULL;

		if (p)
			delete p;
	}
}

void AddressBookIndexScreen::clicked( pair<string, string> entryArg)
{
	keepgoing=false;
	addressBook->currentRecord=entryArg;
	myServer::nextScreen= &addressbookScreen;
	myServer::nextScreenArg=addressBook;
}

void addressbookIndexScreen(void *vp)
{
	AddressBook *a=(AddressBook *)vp;

	vector< pair<string, string> > folder_indexArray;

	{
		list< pair<string, string> > folder_index;

		a->getIndex(folder_index);

		folder_indexArray.reserve(folder_index.size());

		folder_indexArray.insert(folder_indexArray.end(),
					 folder_index.begin(),
					 folder_index.end());
	}

	AddressBookIndexScreen screen(mainScreen, a, folder_indexArray);

	screen.requestFocus();
	myServer::eventloop();
}


/////////////////////////////////////////////////////////////////////////////
//
// List all address books
//

class ListAddressBookScreen : public CursesContainer ,
			      public CursesKeyHandler {

	CursesLabel title;
	CursesDialog listDialog;

	class addressbookButton : public CursesButton {

		ListAddressBookScreen *parent;

	public:
		AddressBook *addressBook;
	private:
		std::list<AddressBook *>::iterator listIterator;

		bool processKeyInFocus(const Key &key);

	public:
		addressbookButton(ListAddressBookScreen *parent,
				  AddressBook *addressBook,
				  list<AddressBook *>::iterator
				  listIterator);

		~addressbookButton();

		void clicked();
	};

	list<addressbookButton *> buttons;

	bool processKey(const Curses::Key &key);
	bool listKeys( vector< pair<string,  string> > &list);

public:
	ListAddressBookScreen(CursesMainScreen *parent);
	~ListAddressBookScreen();

	void requestFocus();

	void deleteAddressBook(list<AddressBook *>::iterator addressBook);
	void moveDown(list<AddressBook *>::iterator addressBook);
	void moveUp(list<AddressBook *>::iterator addressBook);
};

ListAddressBookScreen::addressbookButton
::addressbookButton(ListAddressBookScreen *parentArg,
		    AddressBook *addressBookArg,
		    list<AddressBook *>::iterator listIteratorArg)
	: CursesButton(NULL, addressBookArg->getName()),
	  parent(parentArg),
	  addressBook(addressBookArg),
	  listIterator(listIteratorArg)
{
}

ListAddressBookScreen::addressbookButton::~addressbookButton()
{
}

void ListAddressBookScreen::addressbookButton::clicked()
{
	keepgoing=false;
	myServer::nextScreen= &addressbookIndexScreen;
	myServer::nextScreenArg=addressBook;
}

bool ListAddressBookScreen::addressbookButton::processKeyInFocus(const
								 Key &key)
{
	if (listIterator != AddressBook::addressBookList.end())
	{
		if (key == key_DELETEADDRESSBOOK)
		{
			parent->deleteAddressBook(listIterator);
			return true;
		}

		// Control relative search order:

		if (key == key_DOWN) // CTRL-D
		{
			parent->moveDown(listIterator);
			return true;
		}

		if (key == key_UP) // CTRL-U
		{
			parent->moveUp(listIterator);
			return true;
		}
	}


	return CursesButton::processKeyInFocus(key);
}

ListAddressBookScreen::ListAddressBookScreen(CursesMainScreen *parent)
	: CursesContainer(mainScreen),
	  CursesKeyHandler(PRI_SCREENHANDLER),
	  title(this, _("Address Book List")),
	  listDialog(this)
{
	setRow(4);
	setAlignment(Curses::PARENTCENTER);
	title.setAlignment(Curses::PARENTCENTER);

	listDialog.setRow(2);

	if (AddressBook::defaultAddressBook)
	{
		addressbookButton *bp=
			new addressbookButton(NULL,
					      AddressBook::defaultAddressBook,
					      AddressBook::addressBookList
					      .end());

		if (!bp)
			outofmemory();
		
		try {
			buttons.insert(buttons.end(), bp);
		} catch (...) {
			delete bp;
			LIBMAIL_THROW();
		}

		listDialog.addPrompt(NULL, bp);
		listDialog.addPrompt(NULL, NULL);
	}

	list<AddressBook *>::iterator b=AddressBook::addressBookList.begin(),
		e=AddressBook::addressBookList.end();

	while (b != e)
	{
		AddressBook *p= *b;

		addressbookButton *bp=new addressbookButton(NULL, p, b);

		b++;

		if (!bp)
			outofmemory();

		try {
			buttons.insert(buttons.end(), bp);
		} catch (...) {
			delete bp;
			LIBMAIL_THROW();
		}

		listDialog.addPrompt(NULL, bp);
		listDialog.addPrompt(NULL, NULL);
	}

}

void ListAddressBookScreen::requestFocus()
{
	list<addressbookButton *>::iterator b=buttons.begin(), e=buttons.end();

	while (b != e)
	{
		if ( (*b)->addressBook->iwantFocus)
		{
			(*b)->addressBook->iwantFocus=false;
			(*b)->requestFocus();
			return;
		}
		++b;
	}

	b=buttons.begin();
	e=buttons.end();

	if (b != e)
		(*b)->requestFocus();
}

ListAddressBookScreen::~ListAddressBookScreen()
{
	list<addressbookButton *>::iterator b=buttons.begin(), e=buttons.end();

	while (b != e)
	{
		addressbookButton *bp= *b;

		*b++=NULL;

		if (bp)
			delete bp;
	}
}

bool ListAddressBookScreen::processKey(const Curses::Key &key)
{
	if (key == key_ABORT)
	{
		keepgoing=false;
		myServer::nextScreen= &mainMenu;
		myServer::nextScreenArg=NULL;
		PreviousScreen::previousScreen();
		return true;
	}

	return GlobalKeys::processKey(key, GlobalKeys::ADDRESSBOOKLISTSCREEN,
				      NULL);
}

bool ListAddressBookScreen::listKeys( vector< pair<string, string> > &list)
{
	list.push_back(make_pair(Gettext::keyname(_("DELETEADDRESSBOOK:D")),
				 _("Delete")));
	list.push_back( make_pair(Gettext::keyname(_("MOVEUP:^U")),
				  _("Move up")));
	list.push_back( make_pair(Gettext::keyname(_("MOVEDN:^D")),
				  _("Move down")));

	GlobalKeys::listKeys(list, GlobalKeys::ADDRESSBOOKLISTSCREEN);
	return false;
}


void listaddressbookScreenImport(void *vp)
{
	ListAddressBookScreen screen(mainScreen);

	screen.requestFocus();
	myServer::eventloop();
}

void listaddressbookScreen(void *vp)
{
	AddressBook::importName="";
	AddressBook::importAddr="";

	listaddressbookScreenImport(vp);
}

void ListAddressBookScreen
::deleteAddressBook(list<AddressBook *>::iterator addressBook)
{
	myServer::promptInfo response=
		myServer::prompt(myServer
				 ::promptInfo(Gettext(_("Delete address book \"%1%\"? (Y/N) "))
					      << (*addressBook)->getName())
				 .yesno());

	if (response.abortflag || (string)response != "Y")
		return;

	AddressBook *p= *addressBook;

	p->close();

	AddressBook::addressBookList.erase(addressBook);
	delete p;

	myServer::saveconfig();

	Curses::keepgoing=false;
	myServer::nextScreen= &listaddressbookScreenImport;
	myServer::nextScreenArg=NULL;
}

void ListAddressBookScreen
::moveDown(list<AddressBook *>::iterator ptr)
{
	list<AddressBook *>::iterator ptrNext=ptr;

	++ptrNext;

	if (ptrNext == AddressBook::addressBookList.end())
		return;

	(*ptr)->iwantFocus=true;
	AddressBook::addressBookList.insert(ptr, *ptrNext);
	AddressBook::addressBookList.erase(ptrNext);
	myServer::saveconfig();

	Curses::keepgoing=false;
	myServer::nextScreen= &listaddressbookScreenImport;
	myServer::nextScreenArg=NULL;
}

void ListAddressBookScreen
::moveUp(list<AddressBook *>::iterator ptr)
{
	if (ptr == AddressBook::addressBookList.begin())
		return;

	(*ptr)->iwantFocus=true;

	list<AddressBook *>::iterator ptrPrev=ptr;

	--ptrPrev;

	AddressBook::addressBookList.insert(ptrPrev, *ptr);
	AddressBook::addressBookList.erase(ptr);
	myServer::saveconfig();

	Curses::keepgoing=false;
	myServer::nextScreen= &listaddressbookScreenImport;
	myServer::nextScreenArg=NULL;
}

//////////////////////////////////////////////////////////////////////////////
//
// Quick address book import

class AddressBookTakeScreen : public CursesContainer,
			       public CursesKeyHandler {

	CursesLabel title;
	CursesDialog indexDialog;

	class addressButton : public CursesButton {
		AddressBookTakeScreen *myScreen;

		string name;
		string addr;

	public:
		addressButton(AddressBookTakeScreen *myScreenArg,
			      mail::address addr);
		~addressButton();

		void clicked();
	};

	vector<addressButton *> buttons;

	bool processKey(const Curses::Key &key);

	class indexSort {
	public:
		bool operator()(const mail::address &a,
				const mail::address &b);
	};

public:
	void requestFocus()
	{
		if (buttons.size() > 0)
			buttons[0]->requestFocus();
	}

	AddressBookTakeScreen(CursesMainScreen *,
			       vector< mail::address > &index);
	~AddressBookTakeScreen();
};

AddressBookTakeScreen::AddressBookTakeScreen(CursesMainScreen *parent,
					     vector< mail::address > &indexArg)
	: CursesContainer(parent),
	  CursesKeyHandler(PRI_SCREENHANDLER),
	  title(this, "Take/Import Address"),
	  indexDialog(this)
{
	setRow(4);
	setAlignment(Curses::PARENTCENTER);
	title.setAlignment(Curses::PARENTCENTER);

	indexDialog.setRow(2);
	titleBar->setTitles(_("TAKE ADDRESS"), "");

	buttons.reserve(indexArg.size());

	// Sort addresses to import

	sort(indexArg.begin(), indexArg.end(), indexSort());

	vector< mail::address >::iterator b=indexArg.begin(),
		e=indexArg.end();

	while (b != e)
	{
		addressButton *bb=new addressButton(this, *b);
		b++;

		if (!bb)
			outofmemory();

		try {
			buttons.insert(buttons.end(), bb);
		} catch (...) {
			delete bb;
			LIBMAIL_THROW();
		}

		indexDialog.addPrompt(NULL, bb);
	}
}

AddressBookTakeScreen::~AddressBookTakeScreen()
{
	vector<addressButton *>::iterator b=buttons.begin(), e=buttons.end();

	while (b != e)
	{
		addressButton *p= *b;

		*b++ = NULL;

		if (p)
			delete p;
	}
}

bool AddressBookTakeScreen::processKey(const Curses::Key &key)
{
	if (key == key_ABORT)
	{
		keepgoing=false;
		myServer::nextScreen= &mainMenu;
		myServer::nextScreenArg=NULL;
		PreviousScreen::previousScreen();
		return true;
	}
	return false;
}

AddressBookTakeScreen::addressButton
::addressButton(AddressBookTakeScreen *myScreenArg,
		mail::address addrArg)
	: CursesButton(myScreenArg, addrArg.toString()),
	  name(addrArg.name),
	  addr(addrArg.addr)
{
	setStyle(MENU);
}

AddressBookTakeScreen::addressButton::~addressButton()
{
}

void AddressBookTakeScreen::addressButton::clicked()
{
	AddressBook::importName=name;
	AddressBook::importAddr=addr;

	Curses::keepgoing=false;
	myServer::nextScreen= &listaddressbookScreenImport;
	myServer::nextScreenArg=NULL;
}

vector<mail::address> AddressBook::savedList;

void AddressBook::takeScreen(void *dummyScreen)
{
	AddressBookTakeScreen takeScreen(mainScreen, savedList);

	takeScreen.requestFocus();
	myServer::eventloop();
}

void AddressBook::take(vector<mail::address> &addrListArg)
{
	vector<mail::address> list;

	{
		map<string, mail::address *> addrMap;

		{
			vector<mail::address>::iterator b=addrListArg.begin(),
				e=addrListArg.end();

			while (b != e)
			{
				mail::address &a= *b++;

				if (a.addr.size() == 0)
					continue;

				addrMap.insert(make_pair(a.addr, &a));
			}
		}

		{
			list.reserve(addrMap.size());

			map<string, mail::address *>::iterator
				b=addrMap.begin(),
				e=addrMap.end();

			while (b != e)
			{
				list.push_back( *b->second );
				b++;
			}
		}
	}

	const struct unicode_info &uc= *Gettext::defaultCharset();

	mail::rfc2047::decoder().decode(list, uc, uc);

	if (list.size() == 0)
	{
		statusBar->clearstatus();
		statusBar->status(_("No addresses found."),
				  statusBar->NORMAL);
		statusBar->beepError();
		return;
	}

	savedList=list;

	Curses::keepgoing=false;
	myServer::nextScreen= &takeScreen;
	myServer::nextScreenArg=NULL;
}


void AddressBook::updateAccount(string oldUrl, string newUrl)
{
	if (defaultAddressBook)
		defaultAddressBook->updateAccountChk(oldUrl, newUrl);

	list<AddressBook *>::iterator b=addressBookList.begin(),
		e=addressBookList.end();

	while (b != e)
	{
		(*b)->updateAccountChk(oldUrl, newUrl);
		b++;
	}
}

void AddressBook::updateAccountChk(string oldUrl, string newUrl)
{
	if (url == oldUrl)
		url=newUrl;
}


/////////////////////////////////////////////////////////////////////////////

bool AddressBookTakeScreen::indexSort::operator()(const mail::address &a,
						  const mail::address &b)
{
	int n=a.name.compare(b.name);

	if (n)
		return n < 0;

	return a.addr.compare(b.addr) < 0;
}
