#include "config.h"
#include "curses/cursesfield.H"
#include "curses/cursesstatusbar.H"
#include "cursesaddresslist.H"
#include "gettext.H"
#include "libmail/mail.H"
#include "libmail/rfcaddr.H"
#include "rfc822/rfc822.h"
#include <unistd.h>
#include <errno.h>

using namespace std;

extern CursesStatusBar *statusBar;

//////////////////////////////////////////////////////////////////////////


class CursesAddressList::AddressField : public CursesField {

public:
	CursesAddressList *parent;
	size_t rowNum;

	AddressField(CursesAddressList *parentArg,
		     size_t currentWidth,
		     size_t rowNumArg);
	~AddressField();

	void focusLost();

	bool processKeyInFocus(const Key &key);
};

CursesAddressList::AddressField::AddressField(CursesAddressList *parentArg,
					      size_t currentWidth,
					      size_t rowNumArg)
	: CursesField(parentArg, currentWidth),
	  parent(parentArg),
	  rowNum(rowNumArg)
{
	setUnderlined(false);
}

CursesAddressList::AddressField::~AddressField()
{
}

void CursesAddressList::AddressField::focusLost()
{
	parent->addressFocusLost(rowNum);
}

bool CursesAddressList::AddressField::processKeyInFocus(const Key &key)
{
	return CursesField::processKeyInFocus(key);
}

///////////////////////////////////////////////////////////////////////////

CursesAddressList::CursesAddressList(CursesContainer *parent)
	: CursesContainer(parent), width(parent->getWidth())
{
}

void CursesAddressList::setWidth(int w)
{
	if (w > 0)
	{
		width=w;

		// All AddressFields have the same width

		vector<AddressField *>::iterator b, e;

		b=fields.begin();
		e=fields.end();

		while (b != e)
		{
			AddressField *f= *b++;

			if (f)
				f->setWidth(w);
		}
	}
}

int CursesAddressList::getWidth() const
{
	return width;
}

CursesAddressList::~CursesAddressList()
{
	clear(false);
}

void CursesAddressList::clear(bool doErase)
{
	if (doErase) // Also erase visually.
	{
		vector<AddressField *>::iterator b=fields.begin(),
			e=fields.end();

		while (b != e)
			(*b++)->erase();
	}

	vector<AddressField *>::iterator b=fields.begin(), e=fields.end();

	while (b != e)
	{
		AddressField *p= *b;

		*b++=NULL;

		if (p)
			delete p;
	}
	fields.clear();
}

void CursesAddressList::addressFocusLost(size_t rowNum)
{
	validate(rowNum, fields[rowNum]->getText(), false);
}
#if 0
void CursesAddressList::addressEnter(size_t rowNum)
{
	size_t oldRowCount=fields.size();

	validate(rowNum, fields[rowNum]->getText(), false);

	size_t newRowCount=fields.size();

	if (newRowCount >= oldRowCount)
	{
		rowNum += newRowCount - oldRowCount;

		fields[rowNum]->requestFocus();
		fields[rowNum]->processKeyInFocus(Key(Key::END));
	}
}
#endif

// Validate new contents of one of the AddressFields.

bool CursesAddressList::validate(size_t rowNum,  // Changed AddressField
				 string addresses, // What it has
				 bool ignoreError)
{
	if (rowNum >= fields.size())
		return true;

	size_t errIndex;

	vector<mail::address> addrvec;

	if (!mail::address::fromString(addresses, addrvec, errIndex))
	{
		statusBar->clearstatus();
		statusBar->status(strerror(errno));
		statusBar->beepError();
		return false;
	}

	if (errIndex <= addresses.size())
	{
		vector<wchar_t> buf;

		// Figure out the cursor position

		string s=addresses.substr(0, errIndex);

		mbtow(s.c_str(), buf);

		fields[rowNum]->requestFocus();
		fields[rowNum]->setCursorPos(buf.size());
		statusBar->clearstatus();
		statusBar->status(_("Syntax error"));
		statusBar->beepError();
		return false;
	}

	{
		vector<mail::address> newAddrVec;

		if (addressLookup(addrvec, newAddrVec))
			addrvec=newAddrVec;
	}

	if (addrvec.size() == 0) // This field is empty.
	{
		if (fields.size() <= 1) // Leave one field standing.
			return true;

		// Adjust the array.

		fields[rowNum]->erase();
		delete fields[rowNum];

		fields[rowNum]=NULL;

		fields.erase(fields.begin() + rowNum);

		// Repoint the remaining fields.

		size_t i;

		for (i=rowNum; i<fields.size(); i++)
			if (fields[i])
				fields[i]->rowNum=i;

		for (i=rowNum; i<fields.size(); i++)
			if (fields[i])
				fields[i]->setRow(i);

		// Make sure everything but the last field has a trailing
		// comma.

		i=fields.size()-1;

		string s=fields[i]->getText();

		if (s.size() > 0 && s[s.size()-1] == ',')
			fields[i]->setText(s.substr(0, s.size()-1));

		resized();
		return true;
	}

	if (addrvec.size() > 1)
	{
		// If there's more than one address, insert more fields.

		fields.insert(fields.begin() + rowNum + 1,
			      addrvec.size()-1, NULL);

		size_t i=1;

		try {
			while (i < addrvec.size())
			{
				if ((fields[i+rowNum]=
				     new AddressField(this,
						      getWidth(),
						      i + rowNum))  == NULL)
					outofmemory();
				fields[i+rowNum]->setRow(i+rowNum);
				i++;
			}
		} catch (...) {

			while (i)
			{
				--i;
				delete fields[i];
				fields[i]=NULL;
			}
			fields.erase(fields.begin() + rowNum+1,
				     fields.begin() + rowNum + addrvec.size());
			LIBMAIL_THROW();
		}

		for (i=fields.size(); i > rowNum + addrvec.size(); )
		{
			--i;
			if (fields[i])
				fields[i]->setRow(i);
		}
		resized();
	}

	// Common code to repopulate the changed AddressFields.

	vector<mail::address>::iterator b, e;
	b=addrvec.begin();
	e=addrvec.end();

	while (b != e)
	{
		string s= (*b++).toString();

		if (rowNum + 1 < fields.size())
			s += ","; // All but last addy has a trailing comma

		if (fields[rowNum])
			fields[rowNum++]->setText(s);
	}
	return true;
}

void CursesAddressList::resized()
{
	CursesContainer::resized();

	int w=getWidth();

	vector<AddressField *>::iterator b=fields.begin(), e=fields.end();

	while (b != e)
	{
		if (*b)
			(*b)->setWidth(w);
		b++;
	}
}

void CursesAddressList::requestFocus()
{
	if (fields.size() > 0 && fields[0])
		fields[0]->requestFocus();
}

bool CursesAddressList::isFocusable()
{
	return false; // Child processes are focusable, really.

	//	return fields.size() > 0 && fields[0];
}

bool CursesAddressList::validateAll()
{
	size_t i;

	for (i= fields.size(); i > 0; )
	{
		--i;
		if (fields[i])
			if (!validate(i, fields[i]->getText(), false))
			{
				fields[i]->requestFocus();
				statusBar->clearstatus();
				statusBar->status(_("Syntax error in this header."));
				statusBar->beepError();
				return false;
			}
	}
	return true;
}

bool CursesAddressList::getAddresses( vector<mail::address> &addressArray)
{
	size_t i;

#if 0
	for (i= fields.size(); i > 0; )
	{
		--i;
		if (fields[i])
			if (!validate(i, fields[i]->getText(), ignoreErr))
				return false;
	}
#endif

	string s="";

	for (i=0; i<fields.size(); i++)
		if (fields[i])
			s += fields[i]->getText();

	size_t dummy;

	return mail::address::fromString(s, addressArray, dummy);
}

void CursesAddressList::setAddresses( vector<mail::address> &addressArray )
{
	clear(true);

	AddressField *p=new AddressField(this, getWidth(), 0);

	if (!p)
		outofmemory();

	try {
		fields.push_back(p);
	} catch (...) {
		delete p;
		LIBMAIL_THROW();
	}
	resized();
	validate(0, mail::address::toString("", addressArray), true);
}

void CursesAddressList::setRow(int r)
{
	erase();
	CursesContainer::setRow(r);
	draw();
}

void CursesAddressList::setCol(int c)
{
	erase();
	CursesContainer::setCol(c);
	draw();
}
